@@ -1,10 +1,10 @@
 ;;; vim-tab-bar.el --- Vim-like tab bar -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2024-2025 James Cherti | https://www.jamescherti.com/contact/
+;; Copyright (C) 2024-2026 James Cherti | https://www.jamescherti.com/contact/
 
 ;; Author: James Cherti
-;; Package-Version: 1.1.0
-;; Package-Revision: 1.1.0-0-g41f57f059bd8
+;; Package-Version: 1.1.4
+;; Package-Revision: 1.1.4-0-g22d2d8a92b8d
 ;; URL: https://github.com/jamescherti/vim-tab-bar.el
 ;; Keywords: frames
 ;; Package-Requires: ((emacs "28.1"))
@@ -52,6 +52,7 @@
 ;;; Code:
 
 (require 'tab-bar)
+(require 'cl-lib)
 
 (defgroup vim-tab-bar nil
   "Non-nil if vim-tab-bar mode mode is enabled."
@@ -59,11 +60,45 @@
   :prefix "vim-tab-bar-"
   :link '(url-link "https://github.com/jamescherti/vim-tab-bar.el"))
 
-(defcustom vim-tab-bar-show-groups nil
-  "Show groups in the tab-bar."
-  :type 'boolean
+(defcustom vim-tab-bar-format-tabs-groups
+  '(tab-bar-format-tabs-groups tab-bar-separator)
+  "The `tab-bar-format' used when `vim-tab-bar-show-groups' is non-nil."
+  :type '(repeat (choice (function-item tab-bar-format-tabs-groups)
+                         (function-item tab-bar-format-tabs)
+                         (function-item tab-bar-separator)
+                         (function-item tab-bar-format-align-right)
+                         (function-item tab-bar-format-global)
+                         function))
   :group 'vim-tab-bar)
 
+(defcustom vim-tab-bar-format-tabs
+  '(tab-bar-format-tabs tab-bar-separator)
+  "The `tab-bar-format' used when `vim-tab-bar-show-groups' is nil."
+  :type '(repeat (choice (function-item tab-bar-format-tabs-groups)
+                         (function-item tab-bar-format-tabs)
+                         (function-item tab-bar-separator)
+                         (function-item tab-bar-format-align-right)
+                         (function-item tab-bar-format-global)
+                         function))
+  :group 'vim-tab-bar)
+
+(defun vim-tab-bar--apply-tab-bar-show-groups (value)
+  "Update the tab-bar display according to `vim-tab-bar-show-groups'.
+If VALUE is non-nil, tab groups are shown; otherwise, only tabs are displayed."
+  (when (bound-and-true-p vim-tab-bar-mode)
+    (setq tab-bar-format (if value
+                             vim-tab-bar-format-tabs-groups
+                           vim-tab-bar-format-tabs))
+    (force-mode-line-update)))
+
+(defcustom vim-tab-bar-show-groups nil
+  "If non-nil, display tab groups in the tab-bar."
+  :type 'boolean
+  :group 'vim-tab-bar
+  :set (lambda (symbol enabled)
+         (set-default symbol enabled)
+         (vim-tab-bar--apply-tab-bar-show-groups enabled)))
+
 (defun vim-tab-bar--default-format-tab-name (name)
   "Return NAME surrounded by spaces."
   (concat " " name " "))
@@ -144,21 +179,53 @@ ungrouped tabs."
          (fg-tab-inactive fg-modeline-inactive)
          (fg-tab-active fg-default)
          (bg-tab-active bg-default))
-    (with-suppressed-warnings ((obsolete tab-bar-new-button-show))
-      (setq tab-bar-new-button-show nil))  ; Obsolete as of Emacs 28.1
-    (setq tab-bar-separator "\u200B")  ; Zero width space to fix color bleeding
-    (setq tab-bar-auto-width nil)
+    (cl-letf (((symbol-function 'force-mode-line-update)
+               (lambda (&rest _) nil)))
+      (with-suppressed-warnings ((obsolete tab-bar-new-button-show))
+        (setq tab-bar-new-button-show nil))  ; Obsolete as of Emacs 28.1
 
-    ;; Using setq in these defcustom. These have :set function, but
-    ;; tab-bar-format will call the same.
-    (setq tab-bar-tab-hints nil)  ; Tab numbers on the left
-    (setq tab-bar-close-button-show nil)
-    (setq tab-bar-tab-name-format-function #'vim-tab-bar--name-format-function)
-    (setq tab-bar-tab-group-format-function #'vim-tab-bar--group-format-function)
+      ;; Using the zero-width space (\u200B) prevents background color bleeding
+      ;; between tabs in graphical frames, ensuring a clean visual interface.
+      ;;
+      ;; While this character is rendered as a highlighted "unprintable" glyph
+      ;; in terminal mode, this trade-off is accepted because preserving the
+      ;; visual integrity of the GUI is more important than the highlighting
+      ;; artifact in the terminal.
+      ;;
+      ;; Consequently, the separator is only cleared when no graphical frames
+      ;; exist in the current session.
+      (if (display-graphic-p)
+          ;; Use the zero-width space \u200B as a separator, you ensure that
+          ;; graphical frames maintain distinct, clean boundaries between tabs
+          ;; rather than suffering from background color leakage.
+          (setq tab-bar-separator "\u200B")
+        (if (seq-some #'display-graphic-p (frame-list))
+            ;; Use the zero-width space \u200B as a separator, you ensure that
+            ;; graphical frames maintain distinct, clean boundaries between tabs
+            ;; rather than suffering from background color leakage.
+            (setq tab-bar-separator "\u200B")
+          ;; Revert to an empty string when there no graphical frames are
+          ;; present.
+          (setq tab-bar-separator "")))
 
-    (if vim-tab-bar-show-groups
-        (setq tab-bar-format '(tab-bar-format-tabs-groups tab-bar-separator))
-      (setq tab-bar-format '(tab-bar-format-tabs tab-bar-separator)))
+      (setq tab-bar-auto-width nil)
+
+      ;; Using setq in these defcustom. These have :set function, but
+      ;; tab-bar-format will call the same.
+      (setq tab-bar-tab-hints nil)  ; Tab numbers on the left
+      (setq tab-bar-close-button-show nil)
+      (setq tab-bar-tab-name-format-function #'vim-tab-bar--name-format-function)
+      (setq tab-bar-tab-group-format-function #'vim-tab-bar--group-format-function)
+
+      (vim-tab-bar--apply-tab-bar-show-groups vim-tab-bar-show-groups))
+
+    ;; Ensure that any changes to user options that affect the mode line or UI,
+    ;; such as tab-bar formatting and appearance, are immediately reflected by
+    ;; forcing a mode line update. This prevents inconsistencies where the
+    ;; variables have been set programmatically but the display does not yet
+    ;; reflect those changes, guaranteeing that the tab bar and related elements
+    ;; are drawn correctly without waiting for the next automatic redisplay.
+    (force-mode-line-update)
 
     ;; The tab bar's appearance
     (set-face-attribute 'tab-bar frame
@@ -268,23 +335,29 @@ ungrouped tabs."
                         :strike-through 'unspecified
                         :stipple 'unspecified)
 
-    (when (and (display-graphic-p)
-               (not (daemonp)))
+    ;; The condition (display-graphic-p frame) fixes: entered--Lisp error:
+    ;; (error "Invalid face box" :line-width 3 :color unspecified :style nil)
+    (when (display-graphic-p frame)
       (set-face-attribute
        'tab-bar frame
        :box `(:line-width 3 :color ,bg-tab-inactive :style nil))
+
       (set-face-attribute
        'tab-bar-tab-inactive frame
        :box `(:line-width 3 :color ,bg-tab-inactive :style nil))
+
       (set-face-attribute
        'tab-bar-tab frame
        :box `(:line-width 3 :color ,bg-tab-active :style nil))
+
       (set-face-attribute
        'tab-bar-tab-ungrouped frame
        :box `(:line-width 3 :color ,bg-tab-inactive :style nil))
+
       (set-face-attribute
        'tab-bar-tab-group-inactive frame
        :box `(:line-width 3 :color ,bg-tab-inactive :style nil))
+
       (set-face-attribute
        'tab-bar-tab-group-current frame
        :box `(:line-width 3 :color ,bg-tab-inactive :style nil)))))
@@ -293,6 +366,10 @@ ungrouped tabs."
   "Run `vim-tab-bar--apply' after `load-theme'."
   (vim-tab-bar--apply))
 
+(defun vim-tab-bar--server-after-make-frame-hook ()
+  "Apply config and remove the function from `server-after-make-frame-hook'."
+  (vim-tab-bar--apply))
+
 ;;;###autoload
 (define-minor-mode vim-tab-bar-mode
   "Emulate the Vim tab bar.
@@ -303,6 +380,9 @@ visual consistency with the currently active theme's color scheme."
   :group 'vim-tab-bar
   (if vim-tab-bar-mode
       (progn
+        (when (daemonp)
+          (add-hook 'server-after-make-frame-hook
+                    #'vim-tab-bar--server-after-make-frame-hook))
         (vim-tab-bar--apply)
         (advice-add 'load-theme :after #'vim-tab-bar--run-after-load-theme-hook)
         (tab-bar-mode 1))
